/** 
 @file activity-handler.js
 
 @brief This file contain activity handler implementation which handles 
        activity on the sonos device or group. Each sonos has 3 states:
        - actualState, current active
        - lastState, previous 
        - blinkState, special action aka doorbell or custome events
        
        state switch occur usually occur if upnp command PLAY appear.
        other upnp commmands: MUTE/UN-MUTE, SET/INC/DEC VOLUME, PAUSE/RESUME
        are skipped.
        on UNDO or time elapsed active state ends and new state becomes active again:
		- if blink state was active, then actual state is used
		- if actual state was active, then last state is used
        blinkState is special and overrides actual state which is put onhold
		, waiting for the blinkState to end.
		
		Each state holds a restore data part (sonos device current state snapshot) and 
		are used to restore the sonos after state switch.
		
		There are two possibilities of state switch:
		- actual and last state switch, in this case sonos device state is stored
		- blink and actual state switch, in this case sonos device state is stored
		- blink active, actual arrives, in this case blink remains active and actual 
		  is put onhold, only activity id is stored and returned back later when blink 
		  state is invalid.
		
		OTHER: If user is directly changing the sonos, bypassing the dSS,
	                then different musics could be played
		       New activity is ignored if the same activity is already active.
        	   TODO If blink activity is active and another blink different
        	        activity appears then
	           Same activity can have a timeout flag and an undo. The first one
	                prevails, and others for the same activity are ignored.
	                
	           If activity has dontResume flat set then it mean the previos state
	           will not be resumed.
		             
		
 @author: mobile2DS

 @date: 12.1.2014
*/


// Global constants

var ACTUAL_STATE_NAME = "actual_state";
var LAST_STATE_NAME = "last_state";
var BLINK_STATE_NAME = "blink_state";

var INVALID_ACTIVITY_ID = -1;

var UPDATE_COMPLETE_STATE = 0;
var UPDATE_STATE_DATA_ONLY = 1;
var UPDATE_RESTORE_DATA_ONLY = 2;

var GET_COMPLETE_STATE = 0;
var GET_STATE_DATA_ONLY = 1;
var GET_RESTORE_DATA_ONLY = 2;

var STATE_INVALID = 0;
var STATE_INIT = 1;
var STATE_VALID = 2;

var EVENT_TYPE_CALL_SCENE = 0;
var EVENT_TYPE_CUSTOM = 1;
// Fill in any additional event types


/**
 @brief Creates more states (actual, last, ..) for given
        device or group in the property tree.
        
 @param 
  
 @return
  
*/
function createOneState(strBasePath, nItemId, strStateName) {
    lInfo.logln("createOneState: " + strStateName);

    try {
        var sub_par_path = nItemId + '/' + strStateName;
    
        // 0 - has data from actual state, 1 - has data from blink state
        storeSubParameter(strBasePath
                          , sub_par_path + '/name'
                          , strStateName);
        storeSubParameter(strBasePath
                          , sub_par_path + '/valid'
                          , STATE_INVALID);
        storeSubParameter(strBasePath
                          , sub_par_path + '/upnp_cmd'
                          , COMMAND_NONE);
        storeSubParameter(strBasePath
                          , sub_par_path + '/event_id'
                          , 0);
        storeSubParameter(strBasePath
                          , sub_par_path + '/event_type'
                          , 0);
        storeSubParameter(strBasePath
                          , sub_par_path + '/can_resume'
                          , false);
        storeSubParameter(strBasePath
                          , sub_par_path + '/activity_id'
                          , INVALID_ACTIVITY_ID);
                          
       storeSubParameter(strBasePath, sub_par_path + '/valid_restore', 0);
       storeSubParameter(strBasePath, sub_par_path + '/id', 0);
       storeSubParameter(strBasePath, sub_par_path + '/playerIP', 0);
       storeSubParameter(strBasePath, sub_par_path + '/activeAction', 0);
       storeSubParameter(strBasePath, sub_par_path + '/isRadio', 0);
       storeSubParameter(strBasePath, sub_par_path + '/volumeDef', 0);
       storeSubParameter(strBasePath, sub_par_path + '/volumeSlaves', "");
       storeSubParameter(strBasePath, sub_par_path + '/playMode', 0);
       storeSubParameter(strBasePath, sub_par_path + '/trackListPos', 0);
       storeSubParameter(strBasePath, sub_par_path + '/trackPlayPos', 0);
       storeSubParameter(strBasePath, sub_par_path + '/volumeReset', 0);
       storeSubParameter(strBasePath, sub_par_path + '/playUrl', 0);
       storeSubParameter(strBasePath, sub_par_path + '/activeQueueID', 0);
    }
    catch (e) {
        lInfo.logln("Error by creating one state: " + e);
        for ( var i in e) {
            lInfo.logln(i + ': ' + e[i]);
        }                        
    }        
}


/**
 @brief Creates more states (actual, last, blink, ...) for given
        device or group in the property tree.
        
 @param 
  
 @return
  
*/
function createStates(nId, nIsGroup){
    var have_it = false;
    var item_id = 0;
    var base_path = '';
    
    lInfo.logln("createStates: " + nId);
    
    try {    
        if (nIsGroup == 0) {
            item_id = nId;
            have_it = true;
            base_path = 'sonos-players';
        }
        else {
        }

        if (have_it) {
            createOneState(base_path, item_id, 'actual_state');
        
            createOneState(base_path, item_id, 'last_state');

            createOneState(base_path, item_id, 'blink_state');
                        
            Property.store();
        }
    } 
    catch (e) {
        lInfo.logln("Error by adding states to the device/group: " + e);
        for ( var i in e) {
            lInfo.logln(i + ': ' + e[i]);
        }                        
    }        
}


/**
 @brief Get state data
        
 @param 
  
 @return
  
*/
function getStateData(nId, nIsGroup, strStateName){
    var have_it = false;
    var item_id = 0;
    var base_path = '';
    var state = new Object();
    
    if (nIsGroup == 0) {
        item_id = nId;
        have_it = true;
        base_path = 'sonos-players';
    }
    else {
    }

    var found_it = false;
    if (have_it) {
        var stateNode = Property.getNode(base_path + '/' + item_id + '/' + strStateName);
        if (stateNode != null) {
            state.stateName = stateNode.getChild('name').getValue();
            state.valid = stateNode.getChild('valid').getValue();
            state.eventId = stateNode.getChild('event_id').getValue();
            state.upnpCmd = stateNode.getChild('upnp_cmd').getValue();
            state.eventType = stateNode.getChild('event_type').getValue();
            state.canResume = stateNode.getChild('can_resume').getValue();
            state.activityId = stateNode.getChild('activity_id').getValue();
            state.restoreValid = stateNode.getChild('valid_restore').getValue();
                        
            state.playerID = stateNode.getChild('id').getValue();
            state.playerIP = stateNode.getChild('playerIP').getValue();
            state.activeAction = stateNode.getChild('activeAction').getValue();
            state.isRadio = stateNode.getChild('isRadio').getValue();
            state.volumeDef = stateNode.getChild('volumeDef').getValue();
            state.volumeSlaves = stateNode.getChild('volumeSlaves').getValue();
            state.playMode = stateNode.getChild('playMode').getValue();
            state.trackListPos = stateNode.getChild('trackListPos').getValue();
            state.trackPlayPos = stateNode.getChild('trackPlayPos').getValue();
            state.volumeReset = stateNode.getChild('volumeReset').getValue();
            state.playUrl = stateNode.getChild('playUrl').getValue();      
            state.activeQueueID = stateNode.getChild('activeQueueID').getValue();

            found_it = true;
        }
    }
   
    if (found_it) {
        return state;
    }
    else {
        return null;
    }
}


/**
 @brief Get actual state
        
 @param 
  
 @return
  
*/
function getActualState(nId, nIsGroup){
    return getStateData(nId, nIsGroup, 'actual_state');
}


/**
 @brief Get last state
        
 @param 
  
 @return
  
*/
function getLastState(nId, nIsGroup){
    return getStateData(nId, nIsGroup, 'last_state');
}


/**
 @brief Get blink state
        
 @param 
  
 @return
  
*/
function getBlinkState(nId, nIsGroup){
    return getStateData(nId, nIsGroup, 'blink_state');
}


/**
 @brief Update state data
        
 @param 
  
 @return
  
*/
function updateStateData(nId, nIsGroup, strStateName, oState, nUpdateWhat){
    var have_it = false;
    var item_id = 0;
    var base_path = '';
    var updated = false;
    
	lInfo.logln("updateStateData: " + nId); 
	
    try {
        if (nIsGroup == 0) {
            item_id = nId;
            have_it = true;
            base_path = 'sonos-players';
        }
        else {
        }

        if (have_it) {

        	lInfo.logln("updateStateData: will store changed data into the state");
        
            var stateNode = Property.getNode(base_path + '/' + item_id + '/' + strStateName);
            if (stateNode != null) {
                if (nUpdateWhat != UPDATE_RESTORE_DATA_ONLY) {
                    storeSubParameter(base_path
                                      , item_id + '/' + strStateName + '/valid'
                                      , oState.valid);
                    storeSubParameter(base_path
                                      , item_id + '/' + strStateName + '/upnp_cmd'
                                      , oState.upnpCmd);
                    storeSubParameter(base_path
                                      , item_id + '/' + strStateName + '/event_id'
                                      , oState.eventId);
                    storeSubParameter(base_path
                                      , item_id + '/' + strStateName + '/event_type'
                                      , oState.eventType);
                    storeSubParameter(base_path
                                      , item_id + '/' + strStateName + '/can_resume'
                                      , oState.canResume);
                }
                
                if (nUpdateWhat != UPDATE_STATE_DATA_ONLY) {
                    storeSubParameter(base_path
                                      , item_id + '/' + strStateName+'/activity_id'
                                      , oState.activityId);
                    storeSubParameter(base_path
                                      , item_id + '/' + strStateName+'/valid_restore'
                                      , oState.restoreValid);
                    storeSubParameter(base_path
                                      , item_id + '/' + strStateName+'/id'
                                      , nId);
                    storeSubParameter(base_path
                                      , item_id + '/' + strStateName+'/playerIP'
                                      , oState.playerIP);
                    storeSubParameter(base_path
                                      , item_id + '/' + strStateName+'/activeAction'
                                      , oState.activeAction);
                    storeSubParameter(base_path
                                      , item_id + '/' + strStateName+'/isRadio'
                                      , oState.isRadio);
                    storeSubParameter(base_path
                                      , item_id + '/' + strStateName+'/volumeDef'
                                      , oState.volumeDef);
                    storeSubParameter(base_path
                                      , item_id + '/' + strStateName+'/volumeSlaves'
                                      , oState.volumeSlaves);
                    storeSubParameter(base_path
                                      , item_id + '/' + strStateName+'/playMode'
                                      , oState.playMode);
                    storeSubParameter(base_path
                                      , item_id + '/' + strStateName+'/trackListPos'
                                      , oState.trackListPos);
                    storeSubParameter(base_path
                                      , item_id + '/' + strStateName+'/trackPlayPos'
                                      , oState.trackPlayPos);
                    storeSubParameter(base_path 
                                      , item_id + '/' + strStateName+'/volumeReset'
                                      , oState.volumeReset);
                    storeSubParameter(base_path
                                      , item_id + '/' + strStateName+'/playUrl'
                                      , oState.playUrl);
                    storeSubParameter(base_path
                                      , item_id + '/' + strStateName+'/activeQueueID'
                                      , oState.activeQueueID);               
                }
                                                
				lInfo.logln("State updated");
                
                updated = true;
            }
        }
    }
    catch (e) {
        lInfo.logln("Error by update-ing state: " + e);
        for ( var i in e) {
            lInfo.logln(i + ': ' + e[i]);
        }                        
    }
   
    return updated;
}


/**
 @brief Update actual state data
        
 @param 
  
 @return
  
*/
function storeChanges() {
    // Commented because store is done outside - better performance
    //Property.store();
}

/**
 @brief Update actual state data
        
 @param 
  
 @return
  
*/
function updateActualState(nId, nIsGroup, oState, nUpdateWhat){
    return updateStateData(nId, nIsGroup, 'actual_state', oState, nUpdateWhat);
}


/**
 @brief Update last state data
        
 @param 
  
 @return
  
*/
function updateLastState(nId, nIsGroup, oState, nUpdateWhat){
    return updateStateData(nId, nIsGroup, 'last_state', oState, nUpdateWhat);
}


/**
 @brief Update blink state data
        
 @param 
  
 @return
  
*/
function updateBlinkState(nId, nIsGroup, oState, nUpdateWhat){
    return updateStateData(nId, nIsGroup, 'blink_state', oState, nUpdateWhat);
}


/**
 @brief Check if its blink activity aka doorbell
        
 @param 
  
 @returne
  
*/
function isBlinkActivity(eventId, eventType){
    if ((eventId == 73 /* doorbell */) && (eventType == EVENT_TYPE_CALL_SCENE)) {
        return true;
    }
    return false;
}


/**
 @brief Setup restore data for the caller
        
 @param 
  
 @return
  
*/
function setupRestoreData(refRestoreData, refState) {
    lInfo.logln("setupRestoreData");

    refRestoreData.id = refState.playerID;
    refRestoreData.playerIP = refState.playerIP;
    refRestoreData.activeAction = refState.activeAction;
    refRestoreData.isRadio = refState.isRadio;
    refRestoreData.volumeDef = refState.volumeDef;
    refRestoreData.volumeSlaves = refState.volumeSlaves;
    refRestoreData.playMode = refState.playMode;
    refRestoreData.trackListPos = refState.trackListPos;
    refRestoreData.trackPlayPos = refState.trackPlayPos;
    refRestoreData.volumeReset = refState.volumeReset;
    refRestoreData.playUrl = refState.playUrl;
    refRestoreData.activeQueueID = refState.activeQueueID;
}


/**
 @brief On powerloss or restore, restore the device/group to the 
        actual state
        
 @param 
  
 @return
  
*/
function restoreToActualState(nId, nIsGroup){
/*
    // get actual state for this device/group
    var actual_state = new Object();
    
    actual_state = getActualState(nId, nIsGroup);

    if (actual_state.valid == 1) {
        // return restart data to execute upnp action
        var restart_data = new Object();
        
        setupRestoreData(restart_data, actual_state);
        
        return restart_data;    
    }
    else {
        return null;
    }
*/
}


/**
 @brief Reset (invalidate) device/group states
        
 @param 
  
 @return
  
*/
function resetStates(nId, nIsGroup){
    lInfo.logln("resetStates for player " + nId);
    
    try {
        base_path = 'sonos-players';

        var stateNode = Property.getNode('sonos-players' + '/' + nId);
        if (baseNode != null) {

            lInfo.logln("delete existing states for player " + nId);

            var oArrayNode = baseNode.getChildren();
            for ( var iIndex = 0; iIndex < oArrayNode.length; iIndex++) {
                var oChildNode = oArrayNode[iIndex];
                if (oChildNode.getChild('actual_state') != null) {
                    baseNode.removeChild(oChildNode.getChild('actual_state').getValue());
                }
                
                if (oChildNode.getChild('last_state') != null) {
                    baseNode.removeChild(oChildNode.getChild('last_state').getValue());
                }

                if (oChildNode.getChild('blink_state') != null) {
                    baseNode.removeChild(oChildNode.getChild('blink_state').getValue());
                }
                
                if (oChildNode.getChild('special_state') != null) {
                    baseNode.removeChild(oChildNode.getChild('special_state').getValue());
                }                
            }
        }

        createStates(nId, nIsGroup);
    }
    catch (e) {
        lInfo.logln("Error resetting states: " + e);
        for ( var i in e) {
            lInfo.logln(i + ': ' + e[i]);
        }                        
    }
}


/**
 @brief Store restore data into the proper state
        
 @param 
  
 @return
  
*/
function storeRestartDataIntoState(refRestoreData, refState) {
    lInfo.logln("storeRestartDataIntoState");

    refState.playerID = refRestoreData.id;
    refState.playerIP = refRestoreData.playerIP;
    refState.activeAction = refRestoreData.activeAction;
    refState.isRadio = refRestoreData.isRadio;
    refState.volumeDef = refRestoreData.volumeDef;
    refState.volumeSlaves = refRestoreData.volumeSlaves;
    refState.playMode = refRestoreData.playMode;
    refState.trackListPos = refRestoreData.trackListPos;
    refState.trackPlayPos = refRestoreData.trackPlayPos;
    refState.volumeReset = refRestoreData.volumeReset;
    refState.playUrl = refRestoreData.playUrl;
    refState.activeQueueID = refRestoreData.activeQueueID;
}


/**
 @brief Set the restart data of the actual state
        
 @param restartData - restart data parameters
  
 @return 1 - actual state updated, 0 - could not update the actual state
  
*/
function storeUpnpDeviceStatus(restartData) {
    lInfo.logln("storeUpnpDeviceStatus: " + restartData.id);
	
    var blink_state = getBlinkState(restartData.id, 0);
    var result = false;

    lInfo.logln("storeUpnpDeviceStatus: event id " + restartData.sceneID);
    
    if (blink_state.valid == STATE_INIT) {
        blink_state.valid = STATE_VALID;
        blink_state.restoreValid = 1;

        blink_state.activityId = INVALID_ACTIVITY_ID;

        storeRestartDataIntoState(restartData, blink_state);
      
        if (updateBlinkState(restartData.id, 0
                             , blink_state, UPDATE_COMPLETE_STATE)) {
		    lInfo.logln("storeUpnpDeviceStatus: Added device data to the blink state");
            result = true;
        }
    }
    else {
        var actual_state = getActualState(restartData.id, 0);
    
        if ((actual_state.valid == STATE_INIT) && (actual_state.canResume)) {
            lInfo.logln("storeUpnpDeviceStatus: actual state can resume");
            actual_state.restoreValid = 1;
            actual_state.valid = STATE_VALID;
            
            actual_state.activityId = INVALID_ACTIVITY_ID;
            
            if ((blink_state.valid == STATE_VALID) && (blink_state.restoreValid == 1)) {
                lInfo.logln("storeUpnpDeviceStatus: blink state is active -> should not occur");
                actual_state.valid = STATE_INVALID;
                actual_state.restoreValid = 0;            
                actual_state.activityId = INVALID_ACTIVITY_ID;
                
                updateActualState(nId, 0, actual_state, UPDATE_STATE_DATA_ONLY);
            }
            else {
                storeRestartDataIntoState(restartData, actual_state);
              
                if (updateActualState(restartData.id, 0
                                      , actual_state, UPDATE_COMPLETE_STATE)) {
		            lInfo.logln("storeUpnpDeviceStatus: Added device data to the actual state");
                    result = true;
                }
            }
        } else if (actual_state.valid == STATE_INIT) {
            lInfo.logln("storeUpnpDeviceStatus: actual state cannot resume");
        
            actual_state.valid = STATE_VALID;
            actual_state.restoreValid = 0;            
            actual_state.activityId = INVALID_ACTIVITY_ID;
            
            updateActualState(nId, 0, actual_state, UPDATE_STATE_DATA_ONLY);        
        }
    }
    
    storeChanges();
    
    return result;
}


/**
 @brief Store only the activity for the next action to executed after blink state ends
        
 @param
  
 @return 1 - actual state updated, 0 - could not update the actual state
 
 @remark Blink state is active and actual state is postponed
  
*/
function storeActivityForLater(nId, nIsGroup, nActivityId) {
    lInfo.logln("storeActivityForLater: " + nId);

    var actual_state = getActualState(nId, nIsGroup);
    var blink_state = getBlinkState(nId, nIsGroup);
    var result = false;

    if (actual_state.valid == STATE_INIT) {
        actual_state.valid = STATE_VALID;
        actual_state.restoreValid = 0;
                
        actual_state.activityId = INVALID_ACTIVITY_ID;
        
        if ((blink_state.valid == STATE_VALID) && (blink_state.restoreValid == 1)) {
            // as blink state must end and only then we can play, dont store
            // the restore state of the blink state, but copy the restore data
            // from the blink state and put this restore data into the blink state
            // so blink state will later restore us
            
            lInfo.logln("storeActivityForLater: blink state is active");
            
            lInfo.logln("storeActivityForLater: copy actual state with blink restore data");
            updateActualState(nId, 0, blink_state, UPDATE_RESTORE_DATA_ONLY);
            lInfo.logln("storeActivityForLater: copied device data from blink to actual state");
            
            blink_state.restoreValid = 0;
             
            blink_state.activityId = nActivityId;
                         
            updateBlinkState(nId, 0, blink_state, UPDATE_COMPLETE_STATE);
                        
            result = true;
        }
        else {
            lInfo.logln("storeActivityForLater: blink state is not valid -> should not happen");
            
            actual_state.valid = STATE_INVALID;
            actual_state.restoreValid = 0;            
            actual_state.activityId = INVALID_ACTIVITY_ID;
        }
        
        updateActualState(nId, 0, actual_state, UPDATE_STATE_DATA_ONLY);        
    }
    
    storeChanges();
    
    return result;
}


/**
 @brief Handle new start activity (dS preset or dS activity)
        
 @param nId - sonos device id or group id
 @param nIsGroup - 1 if its a group, 0 - single sonos device
 @param eventId - detected event id
 @param eventType - different event type (callscene, uda, ...)
 @param upnpCmd - upnp cmd
 @param canResume - does this state resumes previous state - true
        , false - not (resume to previous state)
 @return true - can continue executing upnp cmd, false - stop
  
*/
function handleNewActivity(nId, nIsGroup, eventId, eventType
                           , upnpCmd, canResume) {
    lInfo.logln("handleNewActivity: " + nId);

    var result = new Object();
    
    result.executeUpnpCmd = false;
    result.storeCurrDevData = false;
    result.storeActivity = false;

    try {
        // get states for this device/group
        var actual_state = getActualState(nId, nIsGroup);
        var blink_state = getBlinkState(nId, nIsGroup);
        var last_state = getLastState(nId, nIsGroup);

        if (isBlinkActivity(eventId, eventType) || (eventType != EVENT_TYPE_CALL_SCENE)) {
            // A doorbell or custom event usually
			
			lInfo.logln("handleNewActivity: blink activity");

		    if (blink_state.valid == STATE_VALID) {
	            if ((blink_state.eventId == eventId)
    	            && (blink_state.eventType == eventType)
		            && (blink_state.upnpCmd == upnpCmd)) {
               		lInfo.logln("handleNewActivity: overwrite blink state with the same blink activity");		        		
                }
                else {
			        lInfo.logln("handleNewActivity: different blink state already valid -> overwrite");
		        }
	            // Use restore data from previous blink state
	            result.storeCurrDevData = false;
	        }
	        else 
	        {
		        lInfo.logln("handleNewActivity: blink state is valid");

		        blink_state.valid = STATE_INIT;
		        result.storeCurrDevData = true;
	        }
	        
            result.executeUpnpCmd = true;		        

	        blink_state.eventId = eventId;
	        blink_state.upnpCmd = upnpCmd;
	        blink_state.eventType = eventType;
	        blink_state.canResume = canResume;
	        updateBlinkState(nId, nIsGroup, blink_state, UPDATE_STATE_DATA_ONLY);
        }
        else if (upnpCmd == COMMAND_PLAY) {
            var same_activity_in_actual_state = 0;
            var same_activity_in_last_state = 0;

            if ((last_state.valid == STATE_VALID)
                && (last_state.eventId == eventId)
                && (last_state.eventType == eventType)                
                && (last_state.upnpCmd == upnpCmd)) {
			    lInfo.logln("handleNewActivity: new activity is the same in last state");
			    
			    same_activity_in_last_state = 1;
            }
                
            if ((actual_state.valid == STATE_VALID)
                && (actual_state.eventId == eventId)
                && (actual_state.evenType == eventType)
                && (actual_state.upnpCmd == upnpCmd)) {
    			    lInfo.logln("handleNewActivity: overwrite actual state with the same activity");
    			    
    			    same_activity_in_actual_state = 1;
    	    } 
    	    else {
		        lInfo.logln("handleNewActivity: last state become actual state and set new actual state");
	        }
		        
	        if (same_activity_in_actual_state == 0) {
                // Last state is now actual state
                
                if (last_state.valid == STATE_INVALID) {
    		        lInfo.logln("handleNewActivity: last state not valid yet -> update with actual state restore data");
                    updateLastState(nId, nIsGroup
                                    , actual_state, UPDATE_RESTORE_DATA_ONLY);
                }
                
                updateLastState(nId, nIsGroup
                                , actual_state, UPDATE_STATE_DATA_ONLY);
            }

            actual_state.eventId = eventId;
            actual_state.upnpCmd = upnpCmd;
			actual_state.eventType = eventType;
			actual_state.canResume = canResume;      
            actual_state.valid = STATE_INIT;
                            
            if (blink_state.valid == STATE_VALID) {
			    lInfo.logln("handleNewActivity: must store only activity to be executed next");

                result.executeUpnpCmd = false;
                result.storeActivity = true;
            }
            else {
			    lInfo.logln("handleNewActivity: must store also upnp device data if can resume");
			
                result.executeUpnpCmd = true;
                if (actual_state.canResume)
                {
				    if (same_activity_in_actual_state == 0) {
    				    lInfo.logln("handleNewActivity: store restore data")
    				    
    				    result.storeCurrDevData = true;
    				}
    				else {
				        lInfo.logln("handleNewActivity: same activity which can resume");
                        actual_state.valid = STATE_VALID;
    				}
                }
                else {
				    lInfo.logln("handleNewActivity: actual cant resume");
                    actual_state.valid = STATE_VALID;                    
                }
            }
                
            updateActualState(nId, nIsGroup, actual_state, UPDATE_STATE_DATA_ONLY);
        }
        else {
			lInfo.logln("handleNewActivity: unknown action and not handled");
        }
    } 
    catch (e) {
        lInfo.logln("Error handling new activity: " + e);
        for ( var i in e) {
            lInfo.logln(i + ': ' + e[i]);
        }                        
    }

    if ((result.storeCurrDevData == false)
        && (result.storeActivity == false)) {
        storeChanges();
    }

    return result;
}


/**
 @brief Stop the activity (dS preset or dS activity)
        
 @param nId - sonos device id or group id
 @param nIsGroup - 1 if its a group, 0 - single sonos device
  
 @return activity ID or player restart Data
  
*/
function handleStopActivity(nId, nIsGroup, eventId, eventType, isUndo) {
	lInfo.logln("handleStopActivity on event " + eventId + " type " + eventType
	            + ", is undo " + isUndo);

    // get states for this device/group
    var actual_state = getActualState(nId, nIsGroup);
    var last_state = getLastState(nId, nIsGroup);
    var blink_state = getBlinkState(nId, nIsGroup);
    
    var result = new Object();
    
    result.restoreData = null;
    result.activityId = INVALID_ACTIVITY_ID;
    result.performStop = false;
    
    var change_state = false;
    
    if (((isBlinkActivity(eventId, eventType))
        || ((blink_state.eventId == eventId)
            && (blink_state.eventType == eventType)))
        && (blink_state.valid == STATE_VALID)) {
	    lInfo.logln("handleStopActivity on blink state");

        blink_state.valid = STATE_INVALID;
        
        if (blink_state.restoreValid == 1) {
            // return restart data to execute upnp action            
            blink_state.restoreValid = 0;
            
            result.restoreData = new Object();
            
            setupRestoreData(result.restoreData, blink_state);
            			
			lInfo.logln("handleStopActivity return restore data from blink state");
        }
        else if (blink_state.activityId != INVALID_ACTIVITY_ID) {
            result.activityId = blink_state.activityId;
            
            blink_state.activityId = INVALID_ACTIVITY_ID;
			
			lInfo.logln("handleStopActivity return activity id from blink state");
        }
        else {
			lInfo.logln("handleStopActivity blink state has no valid restore or activity");
        }
        
        updateBlinkState(nId, nIsGroup, blink_state, UPDATE_COMPLETE_STATE);        
    }
    else if ((actual_state.eventId == eventId)
             && (actual_state.eventType == eventType) 
             && (actual_state.valid == STATE_VALID)) {

		lInfo.logln("handleStopActivity: undo the actual state");
        actual_state.valid = STATE_INVALID;
        if (blink_state.valid == STATE_VALID) {
			lInfo.logln("handleStopActivity: undo actual state when blink state active");
			
			if (blink_state.activityId != INVALID_ACTIVITY_ID) {
			    // blink state holds the activity for the actual state to be executed
			    // but as we got undo, must again set activity to invalid and
			    // get restore data from the actual state (copied from the blink state
			    // before)
			    blink_state.activityId = INVALID_ACTIVITY_ID;
			    if (blink_state.canResume) {
                    lInfo.logln("handleStopActivity: copy blink state with actual restore data");
                    updateBlinkState(nId, 0, actual_state, UPDATE_RESTORE_DATA_ONLY);
                }
                else {
                    lInfo.logln("handleStopActivity: invalidate blink state activity id");
                    updateBlinkState(nId, 0, blink_state, UPDATE_RESTORE_DATA_ONLY);
                }
			}
			else {
                lInfo.logln("handleStopActivity: actual started before blink and is ending when blink state still active");
			    if (blink_state.canResume) {
                    lInfo.logln("handleStopActivity: copy blink state with actual restore data");
                    updateBlinkState(nId, 0, actual_state, UPDATE_RESTORE_DATA_ONLY);
                }
			}
        }
        else if (actual_state.restoreValid == 1) {
            // return restart data to execute upnp action            
            actual_state.restoreValid = 0;
            
            if (actual_state.canResume)
            {
                result.restoreData = new Object();
                
                setupRestoreData(result.restoreData, actual_state);
                				
			    lInfo.logln("handleStopActivity: return restore data");
			}
			else
			{
			    lInfo.logln("handleStopActivity: resume not performed");
			}
        }
        else if (actual_state.activityId != INVALID_ACTIVITY_ID) {
            result.activityId = actual_state.activityId;
            
            actual_state.activityId = INVALID_ACTIVITY_ID;
			
			lInfo.logln("handleStopActivity: return activity id");
        }
        else {
		    lInfo.logln("handleStopActivity: actual state has no valid restore or activity -> perform stop");
		    result.performStop = true;
        }

        if (!actual_state.canResume) {
            last_state.valid = STATE_INVALID;
        }
        
        updateActualState(nId, nIsGroup
                          , last_state, UPDATE_COMPLETE_STATE);
        last_state.valid = STATE_INVALID;
        updateLastState(nId, nIsGroup
                        , last_state, UPDATE_STATE_DATA_ONLY);
    }
    else if ((last_state.eventId == eventId)
             && (last_state.eventType == eventType)
             && (last_state.valid == STATE_VALID)) {
		lInfo.logln("handleStopActivity: undo the last state");
    
        if (actual_state.valid == STATE_VALID) {
            if (blink_state.valid == STATE_VALID) {
				lInfo.logln("handleStopActivity: undo last state when blink state active");
            }
		    lInfo.logln("handleStopActivity: move restore data only to the actual state");

            updateActualState(nId, nIsGroup, last_state, UPDATE_RESTORE_DATA_ONLY);
        }
        
        last_state.valid = STATE_INVALID;
        
        updateLastState(nId, nIsGroup, last_state, UPDATE_STATE_DATA_ONLY);            
    }
    else {
	    lInfo.logln("handleStopActivity: none state valid -> nothing to do");
    }
    
    storeChanges();
    
    return result;
}


/**
 @brief Handle new undo activity (dS preset or dS activity)
        
 @param nId - sonos device id or group id
 @param nIsGroup - 1 if its a group, 0 - single sonos device
  
 @return activity ID or restartData
  
*/
function handleUndoActivity(nId, nIsGroup, eventId) {
	lInfo.logln("handleUndoActivity " + nId);
    
    return handleStopActivity(nId, nIsGroup, eventId, EVENT_TYPE_CALL_SCENE, 1);
}


/**
 @brief Handle stop sonos action (timed playlist ended playing)
        
 @param nId - sonos device id or group id
 @param nIsGroup - 1 if its a group, 0 - single sonos device
  
 @return activity ID or restartData
  
*/
function handleActivityTimeout(nId, nIsGroup, eventId, eventType) {
    // get states for this device/group
    lInfo.logln("handleActivityTimeout: " + nId);

    return handleStopActivity(nId, nIsGroup, eventId, eventType, 0);
}

